// TCP_IP.cpp : Defines the exported functions for the DLL application.
//
/////////////////////////////////////////////////////////////////////////////
// Changelog
//
//    date  | by | description
// ---------+----+-------------------------------------------------------------
//          |    |
//   260116 | SB | Modified the file to handle 255 ports at once and to remove all multithreading
//          |    |


#include "stdafx.h"
#include <windows.h>
#include <tchar.h>

#include <WinIoCtl.h>
#include <InitGuid.h>
#include <setupapi.h>
#include <devguid.h>
#include <regstr.h>
#include <Setupapi.h>

#include "RobotArm.h"

#define WIN32_LEAN_AND_MEAN

#define MAX_COM_PORTS   255

extern "C"
{
  class ComPort
  {
  public:
    ComPort()
    {
      memset(&dcb, 0, sizeof(DCB));
      hPort = INVALID_HANDLE_VALUE;
    }
    DCB dcb;
    HANDLE hPort;
  };

  // Array of ports (to enable using port as index)
  ComPort ComPorts[MAX_COM_PORTS];
  unsigned int PortNum;


/*=----------------------------------------------------------------------=*\
  Name :GetPortDeviceSet
  Desc :Returns a handle to the "Ports" Device Set
  In   :  None
  Out  :  None
  Retn :  HDEVINFO handle
\*=----------------------------------------------------------------------=*/
HDEVINFO GetPortDeviceSet()
{
  HDEVINFO hDevices = INVALID_HANDLE_VALUE;
  DWORD dwGuids = 0;
  BOOL bRes = SetupDiClassGuidsFromName(_T("Ports"), 0, 0, &dwGuids);
  if (dwGuids)
  {
    LPGUID pguids = (LPGUID)HeapAlloc(GetProcessHeap(), 0, sizeof(GUID) * dwGuids);
    if (pguids)
    {
      bRes = SetupDiClassGuidsFromName(_T("Ports"), pguids, dwGuids, &dwGuids);
      if (bRes)
      {
        hDevices = SetupDiGetClassDevs(pguids, NULL, NULL, DIGCF_PRESENT);
      }
      HeapFree(GetProcessHeap(), 0, pguids);
    }
  }
  return hDevices;
}



bool EnumerateComsPorts(bool Refresh)
{
  bool retenum = false;

  if( m_ComsPorts.IsEmpty() || Refresh )
  {
    m_ComsPorts.RemoveAll();
    //set the port array size based on number of ports present
    m_ComsPorts.SetSize(0);
    //obtain device information set for COM ports
    HDEVINFO hDevInfo = GetPortDeviceSet();

    SP_DEVINFO_DATA DeviceInfoData;
    ZeroMemory(&DeviceInfoData, sizeof(SP_DEVINFO_DATA));
    DeviceInfoData.cbSize = sizeof(SP_DEVINFO_DATA);
    DWORD DeviceIndex = 0;
    DWORD CommsIdx = 0;
    bool bPortIsSerial;

    while( SetupDiEnumDeviceInfo(hDevInfo, DeviceIndex, &DeviceInfoData) )
    {
      PortDetails portinfo;
      retenum = true; //seen something
      bPortIsSerial = false;

		  // Enumerate the current device
		  DeviceInfoData.cbSize = sizeof( SP_DEVINFO_DATA);
      // Get the registry key which stores the port settings
      HKEY hDeviceKey = SetupDiOpenDevRegKey( hDevInfo, &DeviceInfoData, DICS_FLAG_GLOBAL, 0, DIREG_DEV, KEY_QUERY_VALUE);
      if( hDeviceKey)
      {
        // Read in the name of the port
        TCHAR pszPortName[20];
        DWORD dwSize = sizeof(pszPortName);
        DWORD dwType = 0;
        if( (RegQueryValueEx(hDeviceKey, _T("PortName"), NULL, &dwType, (LPBYTE) pszPortName, &dwSize) == ERROR_SUCCESS) && (dwType == REG_SZ) )
        {
          // Check if it really is a com port
          if( _tcsnicmp( pszPortName, _T("COM"), 3) == 0)
          {
            int nPortNr = _ttoi( pszPortName + 3 );
            if( nPortNr != 0 )
            {
              portinfo.PortName = pszPortName;
              bPortIsSerial = true;
              //Got the port name, now get a property: Friendly name
              DWORD bfrsize = 0;
              LPTSTR buffer = NULL;
              while (!SetupDiGetDeviceRegistryProperty(hDevInfo, &DeviceInfoData, SPDRP_DEVICEDESC, NULL, (PBYTE)buffer, bfrsize, &bfrsize))
              {
                if (GetLastError() == ERROR_INSUFFICIENT_BUFFER)
                {
                  // Change the buffer size.
                  if (buffer) LocalFree(buffer);
                  // Double the size to avoid problems on
                  // W2k MBCS systems
                  buffer = (LPTSTR)LocalAlloc(LPTR, bfrsize * 2);
                  ZeroMemory(buffer, bfrsize*2);
                }
                else
                {
                  // other errors ... don't care
                  break;
                }
              }

              if( buffer!=NULL && _tcslen(buffer) > 0 )
              {
                //if property found store it in the port details
                portinfo.DevDesc.Format(_T("%s"), buffer);
                LocalFree(buffer);
              }

              //repeat for property: HardwareID
              bfrsize = 0;
              buffer = NULL;
              while (!SetupDiGetDeviceRegistryProperty(hDevInfo, &DeviceInfoData, SPDRP_HARDWAREID, NULL, (PBYTE)buffer, bfrsize, &bfrsize))
              {
                if (GetLastError() == ERROR_INSUFFICIENT_BUFFER)
                {
                  // Change the buffer size.
                  if (buffer) LocalFree(buffer);
                  // Double the size to avoid problems on
                  // W2k MBCS systems
                  buffer = (LPTSTR)LocalAlloc(LPTR, bfrsize * 2);
                  ZeroMemory(buffer, bfrsize*2);
                }
                else
                {
                  // other errors ... don't care
                  break;
                }
              }

              if( buffer!=NULL && _tcslen(buffer) > 0 )
              {
                //if property found store it in the port details
                TCHAR *pBfr = _tcsstr(buffer, _T("VID_"));
                if( pBfr )
                {
                  portinfo.HardwareID.VendorID.Format(_T("%.4s"), pBfr+4);
                  portinfo.HardwareID.VendorID.MakeUpper();

                  pBfr = _tcsstr(buffer, _T("PID_"));
                  if( pBfr )
                  {
                    portinfo.HardwareID.ProductID.Format(_T("%.4s"), pBfr+4);
                    portinfo.HardwareID.ProductID.MakeUpper();
                    portinfo.HardwareID.IdSet = true;
                  }
                }
                LocalFree(buffer);
              }
            }
          }
        }
		  }
		  // Close the key
		  RegCloseKey(hDeviceKey);
      if( bPortIsSerial )
      {
        //store port details in array & grow if necessary (allow for pnp devices)
        m_ComsPorts.SetAtGrow(CommsIdx, portinfo);
        CommsIdx++;
      }
      DeviceIndex++;
    }

    if (hDevInfo)
    {
      SetupDiDestroyDeviceInfoList(hDevInfo);
    }
	}
  else
  {
    //array already populated
    retenum = true;
  }

  return retenum;
}








DECLDIR unsigned char RA2_GetAPI( )
{
    unsigned char retval;
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'A');      //Command Code
    SendCharCOM(PortNum, '\n');     //Command End
    retval = ReadCharCOM(PortNum, 1000, 256);    //Wait for Return
    return retval;
}

DECLDIR unsigned char RA2_HomeAll( )
{
    unsigned char retval;
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'B');      //Command Code
    SendCharCOM(PortNum, '\n');     //Command End
    retval = ReadCharCOM(PortNum, 10000, 256);    //Wait for Home to complete
    return retval;
}

DECLDIR unsigned char RA2_HomeAxis( unsigned char Axis )
{
    unsigned char retval;
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'C');      //Command Code
    SendCharCOM(PortNum, Axis);
    SendCharCOM(PortNum, '\n');     //Command End
    retval = ReadCharCOM(PortNum, 10000, 256);    //Wait for Home to complete
    return retval;
}

DECLDIR void RA2_GetPosition( unsigned int * A, unsigned int * B, unsigned int * C )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'D');      //Command Code
    SendCharCOM(PortNum, '\n');     //Command End
    A[0] = ReadCharCOM(PortNum, 1000, 256);
    A[0] = A[0] | (ReadCharCOM(PortNum, 1000, 256) << 8);
    B[0] = ReadCharCOM(PortNum, 1000, 256);
    B[0] = B[0] | (ReadCharCOM(PortNum, 1000, 256) << 8);
    C[0] = ReadCharCOM(PortNum, 1000, 256);
    C[0] = C[0] | (ReadCharCOM(PortNum, 1000, 256) << 8);
}

DECLDIR void RA2_SetRelay( unsigned char State )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'E');      //Command Code
    SendCharCOM(PortNum, State);
    SendCharCOM(PortNum, '\n');     //Command End
}

DECLDIR void RA2_StorePosition( unsigned char Index )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'F');      //Command Code
    SendCharCOM(PortNum, Index);
    SendCharCOM(PortNum, '\n');     //Command End
}

DECLDIR void RA2_StoreCoords( unsigned char Index, unsigned int A, unsigned int B, unsigned int C )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'G');      //Command Code
    SendCharCOM(PortNum, Index);
    SendCharCOM(PortNum, A & 0xFF);
    SendCharCOM(PortNum, (A >> 8) & 0xFF);
    SendCharCOM(PortNum, B & 0xFF);
    SendCharCOM(PortNum, (B >> 8) & 0xFF);
    SendCharCOM(PortNum, C & 0xFF);
    SendCharCOM(PortNum, (C >> 8) & 0xFF);
    SendCharCOM(PortNum, '\n');     //Command End
}

DECLDIR void RA2_GotoPosition( unsigned char Index )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'H');      //Command Code
    SendCharCOM(PortNum, Index);
    SendCharCOM(PortNum, '\n');     //Command End
}

DECLDIR void RA2_SetMotor( unsigned char Motor, unsigned int Position )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'I');      //Command Code
    SendCharCOM(PortNum, Motor);
    SendCharCOM(PortNum, Position & 0xFF);
    SendCharCOM(PortNum, (Position >> 8) & 0xFF);
    SendCharCOM(PortNum, '\n');     //Command End
}

DECLDIR void RA2_SetMotors( unsigned int A, unsigned int B, unsigned int C )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'J');      //Command Code
    SendCharCOM(PortNum, A & 0xFF);
    SendCharCOM(PortNum, (A >> 8) & 0xFF);
    SendCharCOM(PortNum, B & 0xFF);
    SendCharCOM(PortNum, (B >> 8) & 0xFF);
    SendCharCOM(PortNum, C & 0xFF);
    SendCharCOM(PortNum, (C >> 8) & 0xFF);
    SendCharCOM(PortNum, '\n');     //Command End
}

DECLDIR unsigned char RA2_IsMoving( )
{
    unsigned char retval;
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'K');      //Command Code
    SendCharCOM(PortNum, '\n');     //Command End
    retval = ReadCharCOM(PortNum, 1000, 256);    //Wait for Return
    return retval;
}

DECLDIR void RA2_GetXYZ( signed int * X, signed int * Y, signed int * Z  )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'L');      //Command Code
    SendCharCOM(PortNum, '\n');     //Command End
    X[0] = ReadCharCOM(PortNum, 1000, 256);
    X[0] = X[0] | (ReadCharCOM(PortNum, 1000, 256) << 8);
    Y[0] = ReadCharCOM(PortNum, 1000, 256);
    Y[0] = Y[0] | (ReadCharCOM(PortNum, 1000, 256) << 8);
    Z[0] = ReadCharCOM(PortNum, 1000, 256);
    Z[0] = Z[0] | (ReadCharCOM(PortNum, 1000, 256) << 8);
}

DECLDIR void RA2_SetXYZ( signed int X, signed int Y, signed int Z )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'M');      //Command Code
    SendCharCOM(PortNum, X & 0xFF);
    SendCharCOM(PortNum, (X >> 8) & 0xFF);
    SendCharCOM(PortNum, Y & 0xFF);
    SendCharCOM(PortNum, (Y >> 8) & 0xFF);
    SendCharCOM(PortNum, Z & 0xFF);
    SendCharCOM(PortNum, (Z >> 8) & 0xFF);
    SendCharCOM(PortNum, '\n');     //Command End
}

DECLDIR void RA2_SetGripper( unsigned char Position )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'N');      //Command Code
    SendCharCOM(PortNum, Position);
    SendCharCOM(PortNum, '\n');     //Command End
}

DECLDIR void RA2_GetColour( unsigned char * R, unsigned char * G, unsigned char * B )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'O');      //Command Code
    SendCharCOM(PortNum, '\n');     //Command End
    R[0] = ReadCharCOM(PortNum, 1000, 256);
    G[0] = ReadCharCOM(PortNum, 1000, 256);
    B[0] = ReadCharCOM(PortNum, 1000, 256);
}

DECLDIR void RA2_SetMoveSpeed( unsigned int Speed, unsigned int Increment )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'P');      //Command Code
    SendCharCOM(PortNum, Speed & 0xFF);
    SendCharCOM(PortNum, (Speed >> 8) & 0xFF);
    SendCharCOM(PortNum, Increment & 0xFF);
    SendCharCOM(PortNum, (Increment >> 8) & 0xFF);
    SendCharCOM(PortNum, '\n');     //Command End
}

DECLDIR void RA2_DisableMotors( )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'Z');      //Command Code
    SendCharCOM(PortNum, '\n');     //Command End
}

DECLDIR void RA2_WriteLED( unsigned char LED, unsigned char State )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'a');      //Command Code
    SendCharCOM(PortNum, LED);
    SendCharCOM(PortNum, State);
    SendCharCOM(PortNum, '\n');     //Command End
}

DECLDIR void RA2_RestoreAutoLEDs( )
{
    SendCharCOM(PortNum, ':');      //Command Start
    SendCharCOM(PortNum, 'b');      //Command Code
    SendCharCOM(PortNum, '\n');     //Command End
}





  /*=----------------------------------------------------------------------=*\
     Use :Returns the port number of the first port to respond as a valid FA
  \*=----------------------------------------------------------------------=*/
  DECLDIR unsigned char RA2_ComFindFirst( void )
  {
    unsigned char retval = 1;
    unsigned char found = 0;

    while ((found == 0) && (retval < MAX_COM_PORTS))
    {
      found = RA2_ComQuery(retval);

      if (found == 0)
        retval++;
    }

    if (found == 1)
      return (retval);
    else
      return (0);
  }



/*=----------------------------------------------------------------------=*\
   Use :Check whether a COM port exists in device manager
       :
\*=----------------------------------------------------------------------=*/
	DECLDIR unsigned char RA2_ComQuery( unsigned char Port )
	{
	    //Populate list when Port = 1, then no need to keep re-populating.
       if (Port == 1)
       {
          EnumerateComsPorts(true);
       }
       else
       {
          EnumerateComsPorts(false);
       }

	   TCHAR  combuffer[64];
       _stprintf_s(combuffer, 64, _T("COM%d"), Port);

       for (int x = 0; x < m_ComsPorts.GetSize(); x++)
       {
          if ( m_ComsPorts.GetAt(x).PortName.Compare(combuffer) == 0)
              return true;
       }

       return false;


      //----- No longer called -----
      /*
        if (Port < 10)
        {
          _stprintf_s(combuffer, 64, _T("COM%d"), Port);
        }
        else
        {
          _stprintf_s(combuffer, 64, _T("\\\\.\\COM%d"), Port);
        }

        ComPorts[Port].hPort = CreateFile(combuffer, GENERIC_WRITE|GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, 0, OPEN_EXISTING, 0, NULL);

        if (ComPorts[Port].hPort==INVALID_HANDLE_VALUE)
        {
          return false;
        }
        else
        {
          CloseHandle(ComPorts[Port].hPort);
          ComPorts[Port].hPort = INVALID_HANDLE_VALUE;
        }

        return true;
      */

	}


/*=----------------------------------------------------------------------=*\
   Use :Opens the specified COM port as the stream to use
       :
\*=----------------------------------------------------------------------=*/
	DECLDIR unsigned char RA2_ComOpen( unsigned char Port )
	{
    unsigned char retval = 0;
	  TCHAR  combuffer[64];

    PortNum = Port;

    if (Port < 10)
    {
      _stprintf_s(combuffer, 64, _T("COM%d"), Port);
    }
    else
    {
      _stprintf_s(combuffer, 64, _T("\\\\.\\COM%d"), Port);
    }

    ComPorts[Port].hPort = CreateFile(combuffer, GENERIC_WRITE | GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0);

    ComPorts[Port].dcb.DCBlength=sizeof(DCB);
    if (!GetCommState(ComPorts[Port].hPort, &ComPorts[Port].dcb))
    {
      //error getting state
      retval = 255;
    }

    ComPorts[Port].dcb.BaudRate=CBR_9600;
    ComPorts[Port].dcb.ByteSize=8;
    ComPorts[Port].dcb.StopBits=ONESTOPBIT;
    ComPorts[Port].dcb.Parity=NOPARITY;

    if(!SetCommState(ComPorts[Port].hPort, &ComPorts[Port].dcb))
    {
      //error setting serial port state
      retval = 255;
    }

    COMMTIMEOUTS timeouts={0};
    timeouts.ReadIntervalTimeout=50;
    timeouts.ReadTotalTimeoutConstant=50;
    timeouts.ReadTotalTimeoutMultiplier=10;
    timeouts.WriteTotalTimeoutConstant=50;
    timeouts.WriteTotalTimeoutMultiplier=10;
    if(!SetCommTimeouts(ComPorts[Port].hPort, &timeouts))
    {
      //error occureed. Inform user
      retval = 255;
    }

    return (retval);
  }

  DECLDIR void RA2_SetPort( unsigned char Port )
  {
    PortNum = Port;
  }


/*=----------------------------------------------------------------------=*\
   Use :Closes a previously open COM port
       :
\*=----------------------------------------------------------------------=*/
	DECLDIR unsigned char RA2_ComClose( unsigned char Port )
	{
	  if (ComPorts[Port].hPort != INVALID_HANDLE_VALUE)
	  {
	    CloseHandle(ComPorts[Port].hPort);
	    ComPorts[Port].hPort = INVALID_HANDLE_VALUE; // Invalid handle now
	  }

    return 0;
	}


/*=----------------------------------------------------------------------=*\
   Use :Sets the baud rate for the currently open COM
       :
\*=----------------------------------------------------------------------=*/
	DECLDIR void BaudRateSet( unsigned char Port, unsigned long Baud )
	{
    ComPorts[Port].dcb.DCBlength = sizeof(DCB);
    if (!GetCommState(ComPorts[Port].hPort, &ComPorts[Port].dcb))
    {
      //error getting state
    }

    ComPorts[Port].dcb.BaudRate = Baud;

    ComPorts[Port].dcb.ByteSize=8;
    ComPorts[Port].dcb.StopBits=ONESTOPBIT;
    ComPorts[Port].dcb.Parity=NOPARITY;

    if(!SetCommState(ComPorts[Port].hPort, &ComPorts[Port].dcb))
    {
      //error setting serial port state
    }
	}


/*=----------------------------------------------------------------------=*\
   Use :Reads a character from the COM port
       :Returns read byte, timeout_val if no data (timeout)
\*=----------------------------------------------------------------------=*/
	DECLDIR unsigned int ReadCharCOM( unsigned char Port, unsigned int timeout, unsigned int timeout_val )
	{
	  unsigned char data[2] = {255};
	  DWORD BytesRead= 0;

    COMMTIMEOUTS timeouts={0};
    timeouts.ReadIntervalTimeout=timeout;
    timeouts.ReadTotalTimeoutConstant=timeout;
    timeouts.ReadTotalTimeoutMultiplier=10;
    timeouts.WriteTotalTimeoutConstant=50;
    timeouts.WriteTotalTimeoutMultiplier=10;
    SetCommTimeouts(ComPorts[Port].hPort, &timeouts);

	  ReadFile(ComPorts[Port].hPort, data, 1, &BytesRead, NULL);

	  if (BytesRead)
	    return(data[0]);
	  else
	    return(timeout_val);
	}


/*=----------------------------------------------------------------------=*\
   Use :Transmits a character from the COM port
       :
\*=----------------------------------------------------------------------=*/
	DECLDIR void SendCharCOM( unsigned char Port, unsigned char Data )
	{
	  char data[2] = {0};
	  DWORD BytesRead= 0;

	  data[0] = Data;

	  WriteFile(ComPorts[Port].hPort, data, 1, &BytesRead, NULL);
	}


/*=----------------------------------------------------------------------=*\
   Use :Sets the DTR status for the currently open COM
       :
\*=----------------------------------------------------------------------=*/
	DECLDIR void SetDTRStatePrivate( unsigned char Port, unsigned char state )
	{
	  if (state)
	    ComPorts[Port].dcb.fDtrControl = DTR_CONTROL_ENABLE;
	  else
	    ComPorts[Port].dcb.fDtrControl = DTR_CONTROL_DISABLE;

    if(!SetCommState(ComPorts[Port].hPort, &ComPorts[Port].dcb))
    {
      //error setting serial port state
    }
	}


/*=----------------------------------------------------------------------=*\
   Use :Sets the flow-control bit for the currently open COM
       :
\*=----------------------------------------------------------------------=*/
	DECLDIR void SetFlowControlPrivate( unsigned char Port, unsigned char state )
	{
	  if (state)
	  {
	    ComPorts[Port].dcb.fOutxCtsFlow = TRUE;
	    ComPorts[Port].dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
	  }
	  else
	  {
	    ComPorts[Port].dcb.fOutxCtsFlow = FALSE;
	    ComPorts[Port].dcb.fRtsControl = RTS_CONTROL_DISABLE;
	  }

    if(!SetCommState(ComPorts[Port].hPort, &ComPorts[Port].dcb))
    {
      //error setting serial port state
    }
	}

}